/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: DRfit.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/

/* $Id: DRfit.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
/*----------------------------------------------------------------------- */
/* Routines and parameters for MIDI Delay-Rate fringe fitting             */
/*----------------------------------------------------------------------- */
#include "DRfit.h"
#include "fourg.h"
#include "dft.h"
#include "pfit.h"
#include "interpol.h"
#include "image.h"

/* global parameters */
#define T_FILTER_WIDTH 0 /* width in pixels of time notch filter, 0=only zero*/
#define F_FILTER_WIDTH 3 /* width in pixels of freq notch filter*/

/* constants */
#define CLIGHT 2.997924562e8 /* speed of light */
#define PI    3.141592654    /* pi */
#define TWOPI 6.283185308    /* two pi */

/* work arrays */
static float FTwork[4096], FTwork1[4096], FTwork2[4096];

/* sine/cosine etc arrays for time DFT. */
DFT_info **DFTinfo_array = NULL;
int      DFTinfo_number=0;

/* image_info for delay-rate image output */
image_info *DRimage = NULL;
float      *DRplane = NULL;  /* work array for delay rate plane */

/* internal prototypes */
void DRfind_peak (float **array, int first_delay, int last_delay, int nd, 
		 int first_rate, int last_rate, int nr, float peak_loc[2], 
		float *rms, float *peak);
void DRnotch_filter (float **array, int nf, int nt, float **out);
void DRFT_old_time (float **array, int nf, int nt); /* debug routine */
void DRFT_in_time (float **array, int nf, int nt);
void DRinit_FT_in_time (int nf, int nt, float delta_t,
		      imaging_detector_table_def* detector);
void DRkill_FT_in_time ();
void DRFT_in_freq (float **array, int nf, int nt);
void DRcorrect_delay (float **array, int nf, int nt, float *OPD_off, 
		    imaging_detector_table_def* detector);
void DRavg_in_time (float **array, int nf, int nt);
void DRavg_in_freq (float **array, int nf, float vis[2]);
void DRdemodulate (float **array, int nf, int nt, float *OPD_off, 
		 float target, imaging_detector_table_def* detector);
void DRprint_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title);
image_info* init_DRpix(char *pixfile, fitsfile *inputfptr, 
		       int nd, int nr, float del_d, float del_r, 
		       float del_t, char *obs_date, int *ierr);
void write_DRpix(image_info *image, int nd, int nr, float **array, 
		 float rms, int *ierr);
void close_DRpix(int *ierr);

/*----------------------------------------------------------------------- */
/******************* external functions **********************************/
/*----------------------------------------------------------------------- */

void setup_DRfringe_fit(char *pixfile, fitsfile *inputfptr, 
			int nf, int nt, float time_incr, 
			imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Setup for fit to spectrally dispersed interferometer data using      */
/*   a two dimensional Fourier transform.                                 */
/*   The delay-rate image is initialized if pixfile is non NULL           */
/*   Necessary structures are allocated.                                  */
/*   Inputs:                                                              */
/*      pixfile   name of output Delay-rate image, NULL=>don't write      */
/*      inputfptr cfitsio pointer to input FITS file                      */
/*                used for delay-rate image                               */
/*      nf        Number of frequency channels in spectrum                */
/*      nt        Number of times                                         */
/*      time_incr Time step between samples in sec                        */
/*      detector  imaging_detector table definition structure             */
/*----------------------------------------------------------------------- */
{
  int    mf, mt, jerr;
  float  del_d, del_r, ref_pix;
  double freq, freq0, freq_incr, freq1, freq2, lo_freq, delay, rate;

  /* setup for Time DFTs if not done */
  if (DFTinfo_array == NULL)  
    DRinit_FT_in_time (nf*2, nt*2, time_incr, detector);

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* find low frequency */
  freq1 = freq0 + (0-ref_pix)*freq_incr;
  freq2 = freq0 + (nf-1-ref_pix)*freq_incr;
  if (freq1>freq2) lo_freq = freq2;
  else  lo_freq = freq1;

  /* initialize delay-rate image file if necessary */
  jerr = 0;
  if ((pixfile!=NULL) && (DRimage==NULL)) {
    /* image to be double resolution */
    mf = nf * 2;
    mt = nt * 2;
    del_d = CLIGHT*1.0e6  / (mf*freq_incr);
    del_r = (CLIGHT /lo_freq)  * 1.0e6 / (mt * time_incr);
    DRimage = init_DRpix (pixfile, inputfptr, mf, mt, del_d, del_r, 
			  time_incr, detector->date_obs, &jerr);
    /* if there's a problem - turn off writing */
    if ((jerr!=0) || (DRimage==NULL)) {
      pixfile = NULL;
      DRimage = NULL;
    }
  }
} /* end setup_DRfringe_fit */

void DRfringe_fit_data (float **spectrum, float **big, 
			int nf, int nt, float time_incr,
			float *OPD_off, float *times, int narrow,
			imaging_detector_table_def* detector,
			cal_table_row *cal_row)
/*----------------------------------------------------------------------- */
/*   Fit spectrally dispersed interferometer data using                   */
/*   a two dimensional Fourier transform.                                 */
/*   Also writes delay-rate images to FITS cube if requested.             */
/*   Notes:                                                               */
/*    1) setup_DRfringe_fit must be called before first call to           */
/*       DRfringe_fit_data                                                */
/*    2) shutdown_DRfringe_fit should be called after last call to        */
/*       DRfringe_fit_data                                                */
/*                                                                        */
/*   Steps:                                                               */
/*     1) apply 2D notch filter to suppress non source signals            */
/*        Also zero pad array to big for better resolution                */
/*     2) remove instrumental OPD                                         */
/*     3) FFT in time direction                                           */
/*     4) FT in frequency direction                                       */
/*     5) Estimate location of the peak                                   */
/*   Inputs:                                                              */
/*      spectrum  Measured spectrum  (nf x nt) array of detector values   */
/*      big       Double size work array for zero padded fft              */
/*      nf        Number of frequency channels in spectrum                */
/*      nt        Number of times                                         */
/*      time_incr Time step between samples in sec                        */
/*      OPD_off   OPD offset applied (seconds) per sample                 */
/*      times     Times relative to reference time (sec)                  */
/*      narrow    if true (!=0) restrict the delay and rate windows       */
/*      detector  imaging_detector table definition structure             */
/*   Outputs:                                                             */
/*      cal_row   calibration table structure for result, ignore if NULL  */
/*----------------------------------------------------------------------- */
{
  long i, j, ii, jj, first_delay, first_rate, last_delay, last_rate;
  double freq, freq0, freq_incr, freq1, freq2, lo_freq, delay, rrate;
  float  ref_pix, noise, background, value, sum, target, amp;
  float  peak_val[3], temp, peak[2], rms, faz, vis[2], del_d, del_r;
  int    mf, mt, jerr;

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* find low frequency */
  freq1 = freq0 + (0-ref_pix)*freq_incr;
  freq2 = freq0 + (nf-1-ref_pix)*freq_incr;
  if (freq1>freq2) lo_freq = freq2;
  else  lo_freq = freq1;

  /* notch filter to kill most non source signals in image */
  /* also zero pads to big for better delay, rate resolution */
  DRnotch_filter (spectrum, nf, nt, big); 
  mf = nf * 2;
  mt = nt * 2;

  /* demodulate OPD */
  /* target = 100 microns in sec */
  target = 100.0 / (CLIGHT*1.0e6) ;
  DRdemodulate(big, mf, mt, OPD_off, target, detector);
  
  /* FFT in time */
  if (nt>1) DRFT_old_time (big, mf, mt);

  /* FFT in freq */
  DRFT_in_freq (big, mf, mt);

  /* find peak */
  /* always ignore junk around and below zero delay */
  first_delay = (mf / 2) + 3;

  if (narrow) { /* narrow search */
    last_delay = mf/2 + mf/8; 
    first_rate = mt/2 - mt/8;
    last_rate = mt/2 + mt/8;
  } else { /* wide open search */
    last_delay = mf; 
    first_rate  = 0;
    last_rate = mt;
  } 
  if (first_delay>=mf) first_delay = mf-1;
  if (first_rate>=mt) first_rate = mt-1;
  if (first_rate<0) first_rate = 0;
  if (last_rate>mt) last_rate = mt;

  DRfind_peak (big, first_delay, last_delay, mf, 
	     first_rate, last_rate, mt,
	     peak, &rms, peak_val); 

  /* convert to delay in microns */
  delay =  CLIGHT*1.0e6 * (peak[0] - mf/2) / (mf*freq_incr);

  /* rate channels all set to lowest frequency - convert to micron/s */
  rrate = (CLIGHT /lo_freq)  * 1.0e6 * (peak[1] - mt/2) / (mt * time_incr);

  /* save result if necessary */
  if (cal_row!=NULL) {
    /* convert results to seconds from microns */
    delay *= 1.0e-6 / CLIGHT;
    rrate  *= 1.0e-6 / CLIGHT;
    amp = sqrt (peak_val[1]*peak_val[1] + peak_val[2]*peak_val[2]);
    if (amp<1.0e-5) amp = 1.0;
    cal_row->gain[0] = peak_val[1] / amp;
    cal_row->gain[1] = peak_val[2] / amp;
    cal_row->delay   = delay;
    cal_row->rate    = rrate;
    cal_row->weight  = peak_val[0]/rms;
  } /* end save value */

  /* write delay - rate image if requested */
  if (DRimage) {
    write_DRpix(DRimage, mf, mt, big, rms, &jerr);
    /* if there's a problem - turn off writing */
    if (jerr!=0) {
      DRimage = NULL;
    }
  }

  /* now use this solution to correct and average the data to get phase */
  /* correct delays */
  for (i=0; i<nt; i++) OPD_off[i] += delay + (times[i] * rrate);

  DRcorrect_delay(spectrum, nf, nt, OPD_off, detector);
  
  /* average in time */
  DRavg_in_time (spectrum, nf, nt);

  /* average in freq */
  DRavg_in_freq (spectrum, nf, vis);

  /* save phase (flip sign to make a correction) */
  amp = sqrt (vis[0]*vis[0] + vis[1]*vis[1]);
  if (amp<1.0e-5) amp = 1.0;
  cal_row->gain[0] =  vis[0] / amp;
  cal_row->gain[1] = -vis[1] / amp;

  /* phase */
  faz = 57.296 * atan2(vis[1], vis[0]);

  /* print result */
  delay /= 1.0e-6 / CLIGHT;
  rrate  /= 1.0e-6 / CLIGHT;
  fprintf (stdout, "time = %f  peak =  %f %f, amp = %f snr= %f  delay = %g rate = %g ph = %f\n",
	   cal_row->time*86400.0, peak[0], peak[1], peak_val[0], peak_val[0]/rms, delay, rrate, faz);


} /* end fringe_fit_data */

void shutdown_DRfringe_fit(int *ierr)
/*----------------------------------------------------------------------- */
/*   Shutdown fit to spectrally dispersed interferometer,                 */
/*   output delay-rate image is closed if written and all allocated       */ 
/*   structures are deleted.                                              */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok, else problem with DR image       */
/*----------------------------------------------------------------------- */
{

  if (*ierr) return; /* existing error? */

  /* Close DR image if open */
  close_DRpix(ierr);
  if (*ierr) return;
  
  kill_image_info (DRimage); /* delete global image_info structure */
  DRimage = 0;
  
  /* delete time FFT structures */
  DRkill_FT_in_time ();

  return;
} /* end shutdown_DRfringe_fit */


/*----------------------------------------------------------------------- */
/******************* internal functions **********************************/
/*----------------------------------------------------------------------- */

void DRfind_peak (float **array, int first_delay, int last_delay, int nd, 
		int first_rate, int last_rate, int nr, float peak_loc[2], 
		float *rms, float *peak)
/*----------------------------------------------------------------------- */
/*   Find the maximum magnitude element in a nd x nr complex array        */
/*   The returned position is the local centroid                          */
/*   Inputs:                                                              */
/*      array       Complex array  as (real, imag)                        */
/*      first_delay First delay complex element to consider               */
/*      last_delay  Last delay complex element to consider                */
/*      nd          dimension in delay                                    */
/*      first_rate  First rate complex element to consider                */
/*      last_rate   last rate complex element to consider                 */
/*      nr          dimension in rate                                     */
/*   Output:                                                              */
/*      peak_loc    Location of peak (pixels in delay, rate)              */
/*      rms         RMS of region far from zero delay, rate               */
/*      peak        magnitude, real, imag interpolated at peak            */
/*----------------------------------------------------------------------- */
{
  float val, peak_pos, sumv, sumd, sumr, norm, count=0.0, max_val=-1.0;
  float darray[9][9], sarray[9][9],  carray[9][9], fblank=-1.2345e5;
  float x, y, fit_max;
  double sum = 0;
  int   id, ir, iid, i, i1, i2, j1, j2,  max_delay=-1, max_rate=-1;
  int   jr, jd, rmin, rmax;
  int   realOnly=0; /* do search with real only or amplitude */

  /* debug - print search window*/
  rmin = nr/2 - 5;
  if (rmin<0) rmin = 0;
  rmax = nr/2 + 5;
  if (rmax>nr) rmax = nr;
  /*
  DRprint_array(1, array, nd/2, nd/2+20, rmin, rmax, 
    "Array in Find_peak");
  DRprint_array(2, array, nd/2, nd/2+20, rmin, rmax, 
    "Array in Find_peak");
  */

  /* double loop over allowed portion of the array */
  /* getting sum for RMS and finding peak in subregion */
  for (ir=first_rate; ir<nr; ir++) {
    for (id=first_delay; id<nd; id++) {
      iid = id*2;
      /* index into real array */
      val = array[ir][iid]*array[ir][iid] + 
	    array[ir][iid+1]*array[ir][iid+1];
      sum += val; count++;
      /* search real part? */
      if (realOnly) val =  array[ir][iid];
      /* don't search everywhere */
      if ((id<=last_delay) && (ir<=last_rate)) {
	if (val>max_val) { /* new maximum */
	  max_val = val;
	  max_delay = id;
	  max_rate = ir;
	}
      }
    } /* end loop over delay */
  } /* end loop over rate */

  /* blank arrays */
  for (ir=0; ir<9; ir++) {
    for (id=0; id<9; id++) {
      darray[ir][id] = fblank;
      sarray[ir][id] = fblank;
      carray[ir][id] = fblank;
    }
  }
  /* find local centroid for better estimate */
  i1 = max_delay-4; if (i1<first_delay) i1 = first_delay;
  i2 = max_delay+4; if (i2>=nd) i2 = nd-1;
  j1 = max_rate-4; if (j1<first_rate) j1 = first_rate;
  j2 = max_rate+4; if (j2>=nr) j2 = nr-1;

  /* double loop over allowed portion of the array */
  /* copy amplitudes to work array */
  for (ir=j1; ir<=j2; ir++) {
    for (id=i1; id<=i2; id++) {
      iid = id*2;
      /* index into real array */
      if (!realOnly) { /* amplitude search */
	val = array[ir][iid]*array[ir][iid] + 
	  array[ir][iid+1]*array[ir][iid+1];
	val = sqrt(val);
      } else { /* only real part */
	val = array[ir][iid];
      }
      /* save value */
      jr = ir - max_rate + 4;
      jd = id - max_delay + 4;
      carray[jr][jd] = array[ir][iid];
      sarray[jr][jd] = array[ir][iid+1];
      darray[jr][jd] = val;
      /* delete from rms sum */
      sum -= val*val; count--;
    } /* end loop over delay */
  } /* end loop over rate */

  /* get better positions, peak */
  if (pfit (darray, &fit_max, peak_loc, fblank)) {
    /* fit failed, use peak*/
    peak_loc[0] = (float)max_delay;
    peak_loc[1] = (float)max_rate;
  } else { /* add offsets */
     peak_loc[0] += (float)max_delay;
     peak_loc[1] += (float)max_rate;
  }

  /* normalization factor for FT */
  /* not sure why 8.0 is necessary */
  norm = 8.0 / ((float)nd * (float)nr);
  /* compute RMS */
  if (count>2.0)
    *rms = sqrt(sum/count) * norm;
  else
    *rms = sqrt(max_val) * norm;

  /* finished */
  /*peak[0] = sqrt(max_val) * norm;*/
  /* amplitude */
  peak[0] = fit_max * norm;

  /* real and imaginary */
  x = peak_loc[0] - max_delay + 4;
  y = peak_loc[1] - max_rate + 4;
  peak[1] =  imintp (&carray[0][0], 9, 9, 3, x, y, fblank) * norm;
  peak[2] =  imintp (&sarray[0][0], 9, 9, 3, x, y, fblank) * norm;

  return;
} /* end DRfind_peak */

void DRnotch_filter (float **array, int nf, int nt, float **out)
/*----------------------------------------------------------------------- */
/*   Apply notch filter to a frequency x time array                       */
/*   Uses global parameters:                                              */
/*      T_FILTER_WIDTH  width in pixels of time notch filter, 0=only zero */
/*      F_FILTER_WIDTH  width in pixels of freq notch filter, 0=only zero */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array filtered                                    */
/*      out       Input array filtered and zero padded                    */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, jfreq, ii;
  float scale;

  /* zero fill out */
  for (itime=0; itime<2*nt; itime++) {
    for (ifreq=0; ifreq<4*nf; ifreq++) {
      out[itime][ifreq] = 0.0;
    }
  }

  /* first filtering in time/rate */
  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    jfreq = 2 * ifreq; /* frequency index */
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      FTwork1[ii++] = array[itime][jfreq];
      FTwork1[ii++] = array[itime][jfreq+1];
    }

    /* do the forward FFT */
    fourg (FTwork1, nt, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    FTwork1[0] = 0.0; FTwork1[1] = 0.0; 
    for (j=1; j<=T_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
      ii = (nt - j) * 2;
      FTwork1[ii] = 0.0; FTwork1[ii+1] = 0.0; 
      }

    /* do the reverse FFT */
    fourg (FTwork1, nt, -1, FTwork2);

    /* copy back to array, renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nt;
    for (itime=0; itime<nt; itime++) { 
      array[itime][jfreq]   =  scale * FTwork1[ii++];
      array[itime][jfreq+1] =  scale * FTwork1[ii++];
    }
  } /* end loop over frequency */

  /* Then filtering in freq/delay */
  /* loop over time */
  for (itime=0; itime<nt; itime++) {

    /* do the forward FFT */
    fourg (array[itime], nf, 1, FTwork2);

    /* do the filtering */
    /* Note: fourg leaves the zero cell in the first element */
    array[itime][0] = 0.0; array[itime][1] = 0.0; 
    for (j=1; j<=F_FILTER_WIDTH/2; j++) {
      ii = j * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
      ii = (nf - j) * 2;
      array[itime][ii] = 0.0; array[itime][ii+1] = 0.0; 
      }

    /* do the reverse FFT */
    fourg (array[itime], nf, -1, FTwork2);

    /* renormalize for effects of fourg */
    ii = 0;
    scale = 1.0 / (float)nf;
    for (j=0; j<nf; j++) {
      ii = j * 2;
      array[itime][ii]   =  scale * array[itime][ii];
      array[itime][ii+1] =  scale * array[itime][ii+1];
      /* make copy in out */
      out[itime][ii]   =  array[itime][ii];
      out[itime][ii+1] =  array[itime][ii+1];
    }
  } /* end loop over time */
  /* debug - print array */
  /*DRprint_array(2, array, 0, nf, 0, nt, 
    "Array after DRnotch_filter");*/
} /* end DRnotch_filter */

void DRFT_old_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the time direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array fourier transformed in time (2nd axis)      */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, ii;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    j = 2 * ifreq;
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      FTwork1[ii++] = array[itime][j];
      FTwork1[ii++] = array[itime][j+1];
    }

    /* do the FFT */
    fourg (FTwork1, nt, 1, FTwork2);

    /* rearrange to sensible order (center at center) */
    /* lower half first */
    ii = nt;
    for (itime=0; itime<nt/2; itime++) { 
      array[itime][j]   = FTwork1[ii++];
      array[itime][j+1] = FTwork1[ii++];
    }
    /* then upper */
    ii = 0;
    for (itime=nt/2; itime<nt; itime++) { 
      array[itime][j]   = FTwork1[ii++];
      array[itime][j+1] = FTwork1[ii++];
    }
  } /* end loop in freq slice */
  /* debug - print array */
  /*DRprint_array(array, 0, nf, 0, nt, 
    "Array after DRFT_old_time");*/
} /* end  DRFT_old_time */

void DRFT_in_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the time direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum (second half expected zero padded)*/
/*      nt        Number of times (second half expected zero padded)      */
/*   Output:                                                              */
/*      array     Input array fourier transformed in time (2nd axis)      */
/*                values are sampled at the fringe rate in Hz of the      */
/*                lowest frequency channel                                */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, i, ii;
  float ref_pix, du;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf/2; ifreq++) {
    j = 2 * ifreq;
    /* collect data into FTwork1 */
    ii = 0;
    for (itime=0; itime<nt/2; itime++) { 
      FTwork1[ii++] = array[itime][j];
      FTwork1[ii++] = array[itime][j+1];
    }

    /* do the FFT */
    DFTc2c (DFTinfo_array[ifreq], FTwork1, FTwork2);

    /* copy back (will be twice the size due to zero padding) */
    ii = 0;
    for (itime=0; itime<nt; itime++) { 
      array[itime][j]   = FTwork2[ii++];
      array[itime][j+1] = FTwork2[ii++];
    }
  } /* end loop in freq slice */

  /* debug - print array */
  /*DRprint_array(1, array, 0, nf, 0, nt, 
    "Array after DRFT_in_time");*/
} /* end DRFT_in_time */

void DRinit_FT_in_time (int nf, int nt, float delta_t,
		      imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Set up for time DFTs                                                 */
/*   Creates time DFT structures in global array DFTinfo_array            */
/*   The time FT use a DFT so that the ooutput rate channels in each      */
/*   spectral channel will be the same in s/s rather than Hz.             */
/*   The array to be transformed is expected zero padded in the second    */
/*   half in both time and frequency.                                     */
/*   Inputs:                                                              */
/*      nf        Number of frequency channels                            */
/*      nt        Number of times                                         */
/*      delta_t   Time increment in sec.                                  */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      values in global array array DFTinfo_array.                       */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, i, ii;
  DFT_info *info;
  float ref_pix, du;
  double freq, freq0, freq1, freq2, lo_freq, freq_incr;

  /* create array */
  DFTinfo_array = (DFT_info**)malloc((nf/2)*sizeof(DFT_info*));
  DFTinfo_number = nf/2; /* remember how many */

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* find low frequency */
  freq1 = freq0 + (0-ref_pix)*freq_incr;
  freq2 = freq0 + (nf-1-ref_pix)*freq_incr;
  if (freq1>freq2) lo_freq = freq2;
  else  lo_freq = freq1;

  /* loop over frequency (second half zero padded so don't need FT */
  for (ifreq = 0; ifreq<nf/2; ifreq++) {

    /* prepare for FT, use fringe frequencies at low frequency end  */
    /* scale so that a given delay rate in s/s will be in the same channel
       inregardless of the frequency */
    freq = freq0 + (ifreq-ref_pix)*freq_incr;
    du = (freq/lo_freq) * (1.0 / (nt*delta_t));
    for (i=0; i<nt; i++) {
      FTwork[i]  = (i-nt/2) * du;
      FTwork2[i] = (i-nt/4) * delta_t;
    }

    /* get sines/cosines (last time half zero padded) */
    DFTinfo_array[ifreq] =  
      makeDFT(BACKWARD, (long)nt/2, FTwork2, (long)nt, FTwork); 

  } /* end loop in freq  */
} /* end  DRinit_FT_in_time */

void DRkill_FT_in_time ()
/*----------------------------------------------------------------------- */
/*   Delete global structures  for time DFTs                              */
/*----------------------------------------------------------------------- */
{
  int   ifreq;

  if (DFTinfo_array==NULL) return;

  /* delete structures */
  for (ifreq = 0; ifreq<DFTinfo_number; ifreq++) 
    killDFT (DFTinfo_array[ifreq]);
  free(DFTinfo_array);  /* get rid of main array */

  DFTinfo_array = NULL; /* just in case */
} /* end  DRkill_FT_in_time */

void DRFT_in_freq (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Fourier transform a freq x time array in the freq direction          */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Input array Fourier transfoemed in frequency            */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, jfreq, jj, ii;
  float temp;

  /* debug - print search window*/
  /*DRprint_array(array, nf/2, nf, 0, nt, 
    "Array in DRFT_in_freq before");*/

  /* loop over time */
  for (itime = 0; itime<nt; itime++) {

    /* do the FFT */
    fourg (array[itime], nf, 1, FTwork2);

    /* rearrange to sensible order (center at center) */
    for (ifreq=0; ifreq<nf/2; ifreq++) { 
      jfreq = nf/2 + ifreq;
      ii = ifreq*2;
      jj = jfreq*2;
      /* swap reals */
      temp = array[itime][jj];
      array[itime][jj] = array[itime][ii];
      array[itime][ii] = temp;
      /* swap imaginaries */
      ii++; jj++;
      temp = array[itime][jj];
      array[itime][jj] = array[itime][ii];
      array[itime][ii] = temp;
    } /* end freq loop */
  } /* end loop in time slice */
  /* debug - print search window*/
  /*DRprint_array(array, nf/2-5, nf, 0, nt, 
    "Array in DRFT_in_freq after");*/

} /* end  DRFT_in_freq */

void DRdemodulate (float **array, int nf, int nt, float *OPD_off, 
		 float target, imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Remove instrumental OPD modulation by changing it to target          */
/*   Interpolation done using shift theorem and applying a phase ramp     */
/*   to each spectrum.                                                    */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum (last half zero padded)           */
/*      nt        Number of times (last half zero padded)                 */
/*      OPD_off   OPD offset applied (seconds)                            */
/*      target    target offset (seconds)                                 */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      array     Input array with initial OPD modulation removed         */
/*----------------------------------------------------------------------- */
{ 
  int   itime, ifreq, j, ii, jj;
  float temp;
  double freq, freq0, freq_incr, delay, phase_fact;
  float  ref_pix, phase, sp, cp, re, im, sum;

  /* debug */
  /* subtract mean from each spectrum to kill most of background */
  /* only bother with real part 
  for (itime = 0; itime<nt; itime++) {
    sum = 0.0;
    for (j=0; j<nf; j++) sum += array[itime][j*2];
    sum = sum / nf;
    for (j=0; j<nf; j++) array[itime][j*2] -= sum;
    }*/

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* loop over actual time slice */
  ii = 0;
  for (itime = 0; itime<nt/2; itime++) {
    phase_fact = TWOPI * OPD_off[itime];

    /* loop over frequency applying phase ramp */
    for (ifreq=0; ifreq<nf/2; ifreq++) {
      freq = freq0 + (ifreq-ref_pix)*freq_incr;
      phase = phase_fact * freq;
      sp = sin(phase);
      cp = cos(phase);
      re = array[itime][ifreq*2];
      im = array[itime][ifreq*2+1];
      array[itime][ifreq*2]   = re*cp - im*sp;
      array[itime][ifreq*2+1] = re*sp + im*cp;
    } /* end of loop in freq */
  } /* end loop in time slice */
  /* debug - print search window*/
  /*
  DRprint_array(2, array, 0, 30, 0, nt, 
    "Array after DRdemodulate");
  */
} /* end  DRdemodulate */

void DRcorrect_delay (float **array, int nf, int nt, float *OPD_off, 
		 imaging_detector_table_def* detector)
/*----------------------------------------------------------------------- */
/*   Remove delay from data array                                         */
/*   Interpolation done using shift theorem and applying a phase ramp     */
/*   to each spectrum.                                                    */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*      OPD_off   OPD correction applied (seconds)                        */
/*      detector  imaging_detector table definition structure             */
/*   Output:                                                              */
/*      array     Input array with initial OPD modulation removed         */
/*----------------------------------------------------------------------- */
{ 
  int   itime, ifreq, j, ii, jj;
  float temp;
  double freq, freq0, freq_incr, delay, phase_fact;
  float  ref_pix, phase, sp, cp, re, im, sum;

  /* get info from structures */
  freq0     = detector->crval[0][0]; /* frequency at reference pixel */
  ref_pix   = detector->crpix[0][0]; /* reference pixel 1-rel */
  freq_incr = detector->cd[0][0];    /* frequency spacing */

  /* debug - print array */
  /*
  DRprint_array(1, array, 0, 30, 0, nt, 
	      "Array before DRcorrect_delay");
  DRprint_array(2, array, 0, 30, 0, nt, 
	      "Array before DRcorrect_delay");
  */

  /* loop over actual time slice */
  ii = 0;
  for (itime = 0; itime<nt; itime++) {
    phase_fact = TWOPI * OPD_off[itime];

    /* loop over frequency applying phase ramp */
    for (ifreq=0; ifreq<nf; ifreq++) {
      freq = freq0 + (ifreq-ref_pix)*freq_incr;
      phase = phase_fact * freq;
      sp = sin(phase);
      cp = cos(phase);
      re = array[itime][ifreq*2];
      im = array[itime][ifreq*2+1];
      array[itime][ifreq*2]   = re*cp - im*sp;
      array[itime][ifreq*2+1] = re*sp + im*cp;
    } /* end of loop in freq */
  } /* end loop in time slice */
  /* debug - print array */
  /*
  DRprint_array(1, array, 0, 60, 0, nt, 
	      "Array after DRcorrect_delay");
  DRprint_array(2, array, 0, 60, 0, nt, 
	      "Array after DRcorrect_delay");
  */

} /* end  DRcorrect_delay */

void DRavg_in_time (float **array, int nf, int nt)
/*----------------------------------------------------------------------- */
/*   Doaverage of data in time                                            */
/*   Result multiplied by 2.0 -(is this) to correct for having only reals */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*      nf        Dimension of spectrum                                   */
/*      nt        Number of times                                         */
/*   Output:                                                              */
/*      array     Data averaged in first time slot                        */
/*----------------------------------------------------------------------- */
{
  int   ifreq, itime, j, ii;
  float sumwt, sumre, sumim, isumwt;

  sumwt = (float)nt;    /* all considered good */
  isumwt = 2.0 / sumwt;

  /* loop over frequency */
  for (ifreq = 0; ifreq<nf; ifreq++) {
    j = 2 * ifreq;

    /* sum data */
    sumre = 0.0; sumim = 0.0;
   for (itime=0; itime<nt; itime++) { 
      sumre += array[itime][j];
      sumim += array[itime][j+1];
    }
    /* save in first time slot */
    array[0][j]   = sumre * isumwt;
    array[0][j+1] = sumim * isumwt;

  } /* end loop in freq  */
  /* debug - print array */
  /*
  DRprint_array(1, array, 0, 30, 0, 10, 
      "Array after DRavg_in_time");
  DRprint_array(2, array, 0, 30, 0, 10, 
  "Array after DRavg_in_time");
  */
} /* end  DRavg_in_time */

void DRavg_in_freq (float **array, int nf,  float vis[2])
/*----------------------------------------------------------------------- */
/*   Do average of all data in freq                                       */
/*   Inputs:                                                              */
/*      array     Measured spectrum  (nf x nt) array of detector values   */
/*                data in first time element                              */
/*      nf        Dimension of spectrum                                   */
/*   Output:                                                              */
/*      vis       Data averaged, as (re, im)                              */
/*----------------------------------------------------------------------- */
{
  int   i, j, ii, indx, jndx, n;
  float sumwt, sumre, sumim;

  indx = 0; jndx = 0; 

  /* sum data */
  sumwt = (float)nf; 
  sumre = 0.0; sumim = 0.0;
  for (i=0; i<nf; i++) { 
    j = 2 * i;
    sumre += array[0][j];
    sumim += array[0][j+1];
  }
  
  /* average */
  if (sumwt>0.0) {
    vis[0]  = sumre / sumwt;
    vis[1]  = sumim / sumwt;
  } else { 
    vis[0]  = 1.0;
    vis[1]  = 0.0;
  }
} /* end  DRavg_in_freq */

void DRprint_array (int type, float **array, int first_delay, int last_delay, 
		 int first_rate, int last_rate, char *title)
/*----------------------------------------------------------------------- */
/*   Print selected window of  in 2D (d x r)  complex array               */
/*   Inputs:                                                              */
/*      type        1 = amplitude, 2 = phase                              */
/*      array       Complex array  as (real, imag)                        */
/*      first_delay First delay complex element to print                  */
/*      last_delay  Last  delay complex element to print                  */
/*      first_rate  First rate complex element to print                   */
/*      last_rate   Last rate complex element to print                    */
/*      title       title for display                                     */
/*----------------------------------------------------------------------- */
{
  float val;
  int   id, ir, iid;


  if (type==1) { /* amplitude */
    fprintf (stdout,"Amplitudes x 10**-3:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = array[ir][iid]*array[ir][iid] + 
	  array[ir][iid+1]*array[ir][iid+1];
	val = 0.001*sqrt(val);
	fprintf (stdout," %8.1f", val);
      }
      fprintf (stdout,"\n");
    }
  } else { /* phase */
    fprintf (stdout,"Phases:  %s \n",title);
    for (id=first_delay; id<last_delay; id++) {
      fprintf (stdout,"  if=%3d ",id);
      iid = id*2;
      for (ir=first_rate; ir<last_rate; ir++) {
	/* index into real array */
	val = 57.296 * atan2(array[ir][iid+1], array[ir][iid]);
	fprintf (stdout," %8.1f", val);
      }
      fprintf (stdout,"\n");
    }
  }
    fprintf (stdout,"\n\n");
} /* end print array */


image_info* init_DRpix(char *pixfile, fitsfile *inputfptr,
		       int nd, int nr, 
		       float del_d, float del_r, 
		       float del_t, char *obs_date, int *ierr)
/*----------------------------------------------------------------------- */
/*   If global variable pixfile is set then it is used as the output FITS */
/*   file name, otherwise no file is written.                             */
/*   This file will contain a FITS image cube of the delay-rate functions */
/*   Copies old, writes new history                                       */
/*   Inputs:                                                              */
/*      pixfile     name of output Delay-rate image, NULL=>don't write    */
/*      inputfptr   cfitsio pointer to input FITS file                    */
/*      nd          number of delay pixels                                */
/*      nr          number of rate pixels                                 */
/*      del_d       delay pixel increment                                 */
/*      del_r       rate pixel increment                                  */
/*      del_t       time pixel increment                                  */
/*      obs_date    observing date                                        */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*   Returns                                                              */
/*      image_info pointer, NULL indicated no file to be written.         */
/*----------------------------------------------------------------------- */
{
  char *ctype[3] = {"DELAY","RATE","TIME"};
  float crota[3] = {0.0, 0.0, 0.0};
  double crval[3] = {0.0, 0.0, 0.0};
  int  bp, naxis[3];
  float crpix[3], crinc[3], bscale;
  long n;
  image_info *out=NULL;

  if (*ierr!=0) return NULL;     /* existing error conditions? */
  if (pixfile==NULL) return NULL;

  /* set values */
  bp = 8;  /* save as 8 bit integers */
  /* only half the delay plane */
  naxis[0] = nd/2; naxis[1] = nr; naxis[2] = 1; /* axis dimensions */
  crpix[0] = 1.0; /* reference pixels */
  crpix[1] = nr / 2 + 1.0;
  crpix[2] = 1.0;
  crinc[0] = del_d;
  crinc[1] = del_r;
  crinc[2] = del_t * nr / 2.0;

  fprintf (stdout, "Writing delay-rate images cube to %s\n",pixfile);

  /* make work array for image plane */
  n = nd * nr;
  DRplane = (float*)malloc(n*sizeof(float));
  if (DRplane==NULL) return NULL; /* check allocation */

  /* do it */
  out = make_image (pixfile, bp, naxis, ctype, crval, crpix, crinc, crota, 
		     obs_date, ierr);
  if (out==NULL) return out; /* failed */

  /* add amplitude info */
  bscale = 0.25;  /* units to sigma */
  fits_write_key(out->fptr, TFLOAT, "BSCALE", (void*)(&bscale),
		 "Pixel scaling to sigma", ierr);
  fits_write_key(out->fptr, TSTRING, "BUNIT", (void*)("RMS"),
		 "Pixel units are RMS in image", ierr);
  fits_write_key(out->fptr, TSTRING, "OBJECT", (void*)("MIDI data"),
		 "Fringe fitted simulated data", ierr);
  fits_write_key(out->fptr, TSTRING, "ORIGIN", (void*)("midi fringeDR"),
		 "Creator of FITS file", ierr);

  /* copy history */
  if (inputfptr) copy_history (inputfptr, out->fptr, ierr);

  /* add parameters in history. FUNCTION DOESN'T EXIST */
  /* history(out->fptr, ierr); */

  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d creating output delay-rate image \n", *ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    return NULL;
  }/* end error test */

  return out; /* done */
} /* end init_DRpix */

void write_DRpix(image_info *image, int nd, int nr, float **array, 
		 float rms, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write delay-rate images to open FITS file                            */
/*   Output images are scaled into units  of the rms                  */
/*   Inputs:                                                              */
/*      image       image_info for I/O                                    */
/*      nd          number of delays                                      */
/*      nr          number of rates                                       */
/*      array       Complex array  as (real, imag)                        */
/*      rms         rms in image                                          */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  long i, id, ir, iid, n;
  int  jerr;
  float scale, maxval=-1.0;
  
  if (image==NULL) return;  /* anybody home? */
  if (*ierr!=0) return;     /* existing error conditions? */

  i = 0;
  for (ir=0; ir<nr; ir++) {
    for (id=nd/2; id<nd; id++) {
      iid = id*2;
      DRplane[i++] = 
	sqrt (array[ir][iid]*array[ir][iid] + 
	      array[ir][iid+1]*array[ir][iid+1]);
      if (maxval<DRplane[i-1]) maxval = DRplane[i-1]; /* get max */
    }
  }

  /* scale units of rms, correct units for FFT etc. */
  n = (long)(nd/2) * (long)nr;
  /* scale = 255.0 / maxval;*/
  scale = 1.0 / (rms / (8.0 / (float)(n*2)));
  maxval = 255.0 / 4.0;  /* clipping level - related to BSCALE in header*/
  for (i=0; i<n; i++) {
    DRplane[i] *= scale;                       /* scale */
    if (DRplane[i]>maxval) DRplane[i] = maxval;  /* clip */
  }
  
  image_write(image, DRplane, ierr); /* write it out */
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d writing output delay-rate image \n", *ierr);
    table_print_error_stack();  /* Show any cfitsio error messages */
    jerr = 0;
    close_DRpix (&jerr); /* try to save what's there */
  }/* end error test */
} /* end write_DRpix */


void close_DRpix(int *ierr)
/*----------------------------------------------------------------------- */
/*   Close delay-rate image in FITS file                                  */
/*   Inputs:                                                              */
/*      image       image_info for I/O                                    */
/*   Outputs:                                                             */
/*      ierr        Error code: 0 => ok                                   */
/*----------------------------------------------------------------------- */
{
  if (DRimage==NULL) return; /* anybody home? */
  if (*ierr!=0) return;     /* existing error conditions? */
  image_close(DRimage, ierr);
  if (DRplane!=NULL) free(DRplane);
} /* end close_DRpix */

/* */
